// INCLUDES
#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

// CONFIGURATION BITS
#pragma config FOSC = HS        // Oscillator Selection bits (HS oscillator: High-speed crystal/resonator on RA4/OSC2/CLKOUT and RA5/OSC1/CLKIN)
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled and can be enabled by SWDTEN bit of the WDTCON register)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select bit (MCLR pin function is MCLR)
#pragma config CP = OFF         // Code Protection bit (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Code Protection bit (Data memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Selection bits (BOR enabled)
#pragma config IESO = ON        // Internal External Switchover bit (Internal External Switchover mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enabled bit (Fail-Safe Clock Monitor is enabled)

//#pragma warning disable 752

// DEFINITIONS
#define _XTAL_FREQ  8000000     // Compiler reference
#define LCD_RS      RA2         // LCD Register Select pin
#define LCD_EN      RA1         // LCD Enable pin
#define LCD_DATA    PORTC       // LCD datapins are connected here
#define number      0x30        //
#define data        RB6         // DHT datapin
#define data_dir    TRISB6      // Direction control

// CONSTANTS AND VARIABLES
const char Minutes = 29;                    // Constant holding minutes, easy for debugging
const char Seconds = 59;                    // Constant holding seconds. easy for debugging

unsigned short TOUT = 0, CheckSum, i;       // Used with DHT11
unsigned short T_Byte1, T_Byte2, RH_Byte1, RH_Byte2;    // Used with DHT11

unsigned char temp[] = "00";        // Used for displaying
unsigned char humi[] = "00";        // Used for displaying

unsigned char ch;                   // Chars from RX is received in this
unsigned char gsm_buffer[41];       // Uart char are going in here

unsigned int seconds = Seconds;     // Seconds
unsigned int minutes = Minutes;     // Minutes
unsigned char str[3];               // Used for counting down
unsigned char ip_address[16];       // IP address assign from operator
unsigned char counter[2];           // Used for initial countdown


char char_to_find = 34;             // 34 is the decimal for " in ascil

bool getchar_active = false;        // Used for boolean expressions

// FUNCTIONS
// Function to strobe LCD, "pings" the En-pin
int lcd_strobe(void)
{
    LCD_EN = 1;
    __delay_us(1);
    LCD_EN = 0;
}

// Function to write on the LCD
void lcd_write(unsigned char c)
{
    __delay_ms(1);
    LCD_DATA = ((c >>4) & 0x0f);
    lcd_strobe();
    LCD_DATA = (c & 0x0f);
    lcd_strobe();
}

// Function to reset/clear the LCD
void lcd_clear(void)
{
    LCD_RS = 0;
    lcd_write(0x1);
    __delay_ms(2);
}

// Function to write a string to the LCD
void lcd_puts(const char *s)
{
    LCD_RS = 1;
    while(*s)
        lcd_write(*s++);
}

// Functionto write one char on the LCD
void lcd_putch(char c)
{
    LCD_RS = 1;
    lcd_write(c);
}

// FUnction to place cursor on the LCD
void lcd_goto(unsigned char pos)
{
    LCD_RS = 0;
    lcd_write(0x80+pos);
}

// Functio to initialze the LCD
void lcd_init(void)
{
    char init_value;    
    init_value = 0x3;
  
    LCD_RS = 0;
    LCD_EN = 0;
    
    __delay_ms(15);
      
    LCD_DATA = init_value;
    lcd_strobe();    __delay_ms(10);
    lcd_strobe();    __delay_ms(10);
    LCD_DATA = 2;
    lcd_strobe();   
    lcd_write(0x0e);    // display on, cursor on, blink off
}

// Function to initialize the UART port
void uart_init(void)
{
    TXSTAbits.BRGH  = 1;    // Baud rate select, 0 = low, 1 = hih speed
    TXSTAbits.SYNC  = 0;    // 0 = Async, 1 = sync mode
    TXSTAbits.TX9   = 0;    // 0 = 8-bit trans, 1 = 9-bit trans
    RCSTAbits.CREN  = 1;    // Cont receive, 0 = disable, 1 = enable cont receive.
/*
 * 8MHz Crystal
 * 8000000 / 9600 = 833.3333
 * 833.3333 / 16 = 52.0833
 * 52.0833 - 1 = 51.0833
 * 51.0833 = 51
 * 
 */    
//    SPBRG = 8; //57.6k-n-8-1
    SPBRG = 51; // 9.6k-n-8-1
    
    PIE1bits.RCIE   = 1;    // Usart receive interrupt enable, 0= dis, 1=ena
    RCSTAbits.SPEN  = 1;    // Serial port enable bit, 0 = dis, 1 = enabel
    TXSTAbits.TXEN  = 1;    // Transmit enable bit, = = dis, 1 = enabl   
    return;
}

// Function to initialize the oscillator
void osc_init(void)
{
    OSCCONbits.IRCF = 0b111;  // 8MHz
    OSCCONbits.OSTS = 1;    // Running from Fosc in config
    OSCCONbits.HTS  = 1;    // stable
    OSCCONbits.SCS  = 0;    // Clock source def by Fosc    
    return;
}

// Short awesome welcome message
void short_welcome(void)
{
        lcd_clear();
        lcd_goto(2);
        lcd_puts("Hello World");
        lcd_goto(0x40+3);
        lcd_puts("I'm alive");
        __delay_ms(1500);
        lcd_clear();
        lcd_goto(3);
        lcd_puts("PIC 16F690");
        lcd_goto(0x40);
        lcd_puts("DHT11  LCD  GSM ");
        __delay_ms(1500);        
}

// Function is run each time an interrupt occurs.
void interrupt tc_int(void)
{       
if (RCIE && RCIF)
{
    getchar_active = false;
    RCREG;
}
return;
}

// Function adds \r\n to uart strings
void putch(unsigned char byte)
{
    while (!TXSTAbits.TRMT);
    TXREG = byte;
    if ('\n' == byte)
    {
        while (!TXSTAbits.TRMT);
        TXREG = '\r';
    }
    return;
}

// Function to read char from uart
unsigned char getch()
{
    getchar_active = true;
    while(getchar_active)
        continue;
    return RCREG;
}

// Function to trigger the DHT11
void start_signal()
{
    data_dir = 0;
    data = 0;
    __delay_ms(18);
    data = 1;
    __delay_us(20);
    data_dir = 1;
}

// Function to read one byte from the DHT11
unsigned short read_byte()
{
   unsigned short num = 0, t;
    data_dir = 1;
    for (i=0;i<8;i++)
    {
        while(!data);
        __delay_us(40);
    if(data) num |= 1<<(7-i);
        while(data);
    }
    return num;
}

// Function to check the response from the DHT11
unsigned short check_response()
{
    TOUT = 0;
    TMR2 = 0;
    T2CONbits.TMR2ON = 1;
    while (!data && !TOUT);
    if (TOUT) return 0;
    else
    {
        TMR2 = 0;
        while (data && !TOUT);
        if (TOUT) return 0;
        else
        {
            T2CONbits.TMR2ON = 0;
            return 1;
        }
    }
}

// Function that reads the DHT11 bytes
void read_dht(void)
{
    unsigned short check;
    __delay_ms(1000);
    start_signal();
    check = check_response();
    if (!check)
    {
        printf("No response.\r\n");
    }
    else
    {
        RH_Byte1 = read_byte();
        RH_Byte2 = read_byte();
        T_Byte1 = read_byte();
        T_Byte2 = read_byte();
        CheckSum = read_byte();
        if (CheckSum == ((RH_Byte1+RH_Byte2+T_Byte1+T_Byte2) & 0xFF))
        {            
            temp[0] = T_Byte1 / 10 + 48; 
            temp[1] = T_Byte1 % 10 + 48;
           
            humi[0] = RH_Byte1 / 10 + 48;
            humi[1] = RH_Byte1 % 10 + 48;
        }
        else
        {
            printf("Checksum error!\r\n");
        }
    }
}

// Funcion that show wait and count down from 10
void gsm_wait_for_registration(void)
{
    lcd_clear();
    lcd_puts("Booting GSM");
        for (i=9;i>0;i--)
        {
            lcd_goto(0x40);
            lcd_puts("Wait... ");
            itoa(counter, i , 10);
            lcd_puts(counter);
            lcd_puts("      ");
            __delay_ms(1000);
        }        
}

// Function that get the preferred operator, and show it on the LCD
void gsm_get_operator(void)
{
    int j=0;
    unsigned char pos[3];
    int start, stop;
    printf("AT+CPOL?\r\n");
    ch = getch();       // read /n -> New Line
    ch = getch();       // read /r -> Carrier Return
    ch = getch();       // read the + sign
    if (ch == '+')
    {
        printf("\r\n");               // For debugging  
        lcd_clear();    lcd_goto(0);
        lcd_puts("Operator: ");
        lcd_goto(0x40);        
        for (i=0;i<40;i++)      // For-loop puts received chars in gsm_buffer
        {
            ch = getch();
            gsm_buffer[i] = ch;
        }
        for (i=0;i<strlen(gsm_buffer);i++)  // For-loop search for "
        {
            if (gsm_buffer[i] == char_to_find) // If " is found
            {
                pos[j] = i+1;                   // Remember position
                j++;
            }
        }
        start = pos[0];                         // First occurens of "
        stop = pos[1];                          // Second occurnes of "
        for (i=start;i<stop-1;i++)
        {
            printf("%c", gsm_buffer[i]);  // Prints the Operator -> Debug
            lcd_putch(gsm_buffer[i]);
        }                
        printf("\r\n");                    // debug only
    }    
}

// Function that connects the SIM900 with GPRS
void gsm_connect_gprs(void)
{
    int j=0;
    unsigned char pos[3];
    int start, stop;
    lcd_clear();    lcd_goto(0);
    lcd_puts("Open GPRS.");  
    printf("AT+CGATT=1\r\n");       // Attach to gprs
    __delay_ms(2000);
    printf("AT+SAPBR=1,1\r\n");     // Open a GPRS context
    __delay_ms(2000);
    printf("AT+SAPBR=2,1\r\n");     // To query the GPRS context
    ch = getch();       // read \r
    ch = getch();       // read \n
    ch = getch();       // read the + -sign
    if (ch == '+')
    {

        for(i=0;i<40;i++)
        {
            ch = getch();
            gsm_buffer[i] = ch;
        }
        for (i=0;i<strlen(gsm_buffer);i++)
        {
            if (gsm_buffer[i] == char_to_find)
            {
                pos[j] = i+1;
                j++;
            }
        }
        start = pos[0];
        stop = pos[1];
        lcd_goto(0x40);
        for (i=start;i<stop-1;i++)
        {
            printf("%c", gsm_buffer[i]);        // For debugging
            ip_address[i] = gsm_buffer[i];
            lcd_putch(ip_address[i]);
        }                
        printf("\r\n");                         // For debugging
    } // end if ch == +
    printf("\r\nEnd of function\r\n");          // For debugging
}

// Function that send temperature and humidity to my server.
void gsm_send_data(void)
{
    lcd_clear();    lcd_goto(0);
    lcd_puts("Sending data.");    
    printf("AT+HTTPINIT\r\n");                  // Initialize HTTP
    __delay_ms(1000);
    printf("AT+HTTPPARA=\"URL\",\"http://INSERT_YOUR_SERVERE_HERE/add_temp.php?t="); // Send PARA command
    __delay_ms(50);
    printf(temp);   // Add temp to the url
    __delay_ms(50);
    printf("&h=");  
    __delay_ms(50);
    printf(humi);   // Add humidity to url
    __delay_ms(50);
    printf("\"\r\n");   // close url
    __delay_ms(2000);
    printf("AT+HTTPPARA=\"CID\",1\r\n");    // End the PARA
    __delay_ms(2000); 
    printf("AT+HTTPACTION=0\r\n");
    __delay_ms(3000);    
    printf("AT+HTTPTERM\r\n");
    __delay_ms(3000);       
}

// MAIN PROGRAM
void main(void) {
    TRISA = 0b00000000;     // TRISA output
    PORTA = 0b00000000;     // PORTA low
    
    TRISB = 0b1010;         // RX+TX = in, rest out
    PORTB = 0b0000;         // PORTB low
    
    TRISC = 0b00000000;     // TRISC output
    PORTC = 0b00000000;     // PORTC low
           
    ANSEL = 0;              // Disable the input buffer on ANSEL
    ANSELH= 0;              // Disable the input buffer on ANSELH        
    
    CM1CON0 = 0;            // Comparator C1 control register 0,  disabled
    CM2CON0 = 0;            // Comparator C2 control register 0, disabled
    
    RCIF = 0;               // UART Receive interrupt flag is cleared
    RCIE = 1;               // UART Receive interrupt enable bit is enabled
    
    PEIE = 1;               // Perepheral Interrupt enble bit, 1 = enable
    GIE = 1;                // Global Interrupt Enable bit, 1 = enable
    
    osc_init();             // Initialize the oscillator
    lcd_init();             // Initialize the LCD    
    uart_init();            // Initialize the uart 
    
    short_welcome();        // Show awesome welcome      
    
    gsm_wait_for_registration();    // Show countdown
    
    gsm_get_operator();     // Get and show operator
    
    __delay_ms(2000);
    
    gsm_connect_gprs();     // Connect the SIM900A to GPRS
    
    __delay_ms(2000);
    
    gsm_send_data();        // Send initial data, allways 00 00, haven't read DHT11
    
    __delay_ms(2000);
    
    while(1)                // Do this forever...
    {
    lcd_clear();
    do
    {
        do
        {
                read_dht();     // read the DHT11 sensor

                lcd_goto(0);
                lcd_puts("T: ");        lcd_puts(temp); lcd_putch(0xDF); // 0xDF = degree symbol

                lcd_puts("c RH: ");     lcd_puts(humi); lcd_putch(0x25); // 0x25 = percent symbol
                lcd_goto(0x40);

                lcd_puts("Next: ");
                itoa(str, minutes, 10); lcd_puts(str);
                lcd_puts(":");
                if (seconds < 10)
                {
                    lcd_puts("0");
                }
                itoa(str, seconds, 10);
                lcd_puts(str);          lcd_puts(" ");
                seconds--;
        } while (seconds != 0);
        seconds = Seconds;
        minutes--;
    } while (minutes != -1);
    minutes = Minutes;
    gsm_send_data();        // Now we're sending real data.
    printf(temp);   printf(" - ");  printf(humi);   printf("\r\n"); // Debugging

    }           
    return;
}